"""
The code in this script is heavily borrowed from
https://www.tensorflow.org/text/tutorials/transformer

"""


import tempfile
from pathlib import Path
from typing import Sequence

import numpy as np
import tensorflow as tf
import tensorflow_text as tf_text  # In order to load the tokenizers
from huggingface_hub import snapshot_download

from vina2vi.util import uncased_vina_normalizer, BertDetokenizeHelper


class Tokenizers:
    """
    dict-like tokenizer containing keys `"vi"` and `"vina"`
    """
    # TODO: Implement __init__
    def __init__(self):
        pass

    @classmethod
    def from_pretrained(
        cls, *,
        tokenizers_repo_id: str = "phunc20/vnexpress_uncased_8000_vi_vina_tokenizers",
    ):
        tokenizers_local_dir = Path(snapshot_download(tokenizers_repo_id))
        tokenizers = tf.saved_model.load(tokenizers_local_dir)

        return tokenizers


def positional_encoding(length, depth):
    depth = depth/2

    positions = np.arange(length)[:, np.newaxis]     # (seq, 1)
    depths = np.arange(depth)[np.newaxis, :]/depth   # (1, depth)

    angle_rates = 1 / (10000**depths)         # (1, depth)
    angle_rads = positions * angle_rates      # (pos, depth)

    pos_encoding = np.concatenate(
        [np.sin(angle_rads), np.cos(angle_rads)],
        axis=-1)

    return tf.cast(pos_encoding, dtype=tf.float32)


class PositionalEmbedding(tf.keras.layers.Layer):
    def __init__(self, vocab_size, d_model):
        super().__init__()
        self.d_model = d_model
        self.embedding = tf.keras.layers.Embedding(vocab_size, d_model, mask_zero=True)
        self.pos_encoding = positional_encoding(length=1024, depth=d_model)

    def compute_mask(self, *args, **kwargs):
        return self.embedding.compute_mask(*args, **kwargs)

    def call(self, x):
        length = tf.shape(x)[1]
        x = self.embedding(x)
        # This factor sets the relative scale of the embedding and positonal_encoding
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x = x + self.pos_encoding[tf.newaxis, :length, :]
        return x


class BaseAttention(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__()
        self.mha = tf.keras.layers.MultiHeadAttention(**kwargs)
        self.layernorm = tf.keras.layers.LayerNormalization()
        self.add = tf.keras.layers.Add()


class CrossAttention(BaseAttention):
    def call(self, x, context):
        attn_output, attn_scores = self.mha(
            query=x,
            key=context,
            value=context,
            return_attention_scores=True)

        # Cache the attention scores for plotting later
        self.last_attn_scores = attn_scores

        x = self.add([x, attn_output])
        x = self.layernorm(x)

        return x


class GlobalSelfAttention(BaseAttention):
    def call(self, x):
        attn_output = self.mha(
            query=x,
            value=x,
            key=x,
        )
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


class CausalSelfAttention(BaseAttention):
    def call(self, x):
        attn_output = self.mha(
            query=x,
            value=x,
            key=x,
            use_causal_mask = True,
        )
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


class FeedForward(tf.keras.layers.Layer):
    def __init__(self, d_model, dff, dropout_rate=0.1):
        super().__init__()
        self.seq = tf.keras.Sequential([
            tf.keras.layers.Dense(dff, activation='relu'),
            tf.keras.layers.Dense(d_model),
            tf.keras.layers.Dropout(dropout_rate)
        ])
        self.add = tf.keras.layers.Add()
        self.layer_norm = tf.keras.layers.LayerNormalization()

    def call(self, x):
        x = self.add([x, self.seq(x)])
        x = self.layer_norm(x)
        return x


class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self,*, d_model, num_heads, dff, dropout_rate=0.1):
        super().__init__()

        self.self_attention = GlobalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate,
        )

        self.ffn = FeedForward(d_model, dff)

    def call(self, x):
        x = self.self_attention(x)
        x = self.ffn(x)
        return x


class Encoder(tf.keras.layers.Layer):
    def __init__(self, *, num_layers, d_model, num_heads,
                 dff, vocab_size, dropout_rate=0.1):
        super().__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.pos_embedding = PositionalEmbedding(
            vocab_size=vocab_size, d_model=d_model)

        self.enc_layers = [
            EncoderLayer(d_model=d_model,
                         num_heads=num_heads,
                         dff=dff,
                         dropout_rate=dropout_rate)
            for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x):
        # `x` is token-IDs shape: (batch, seq_len)
        x = self.pos_embedding(x)  # Shape `(batch_size, seq_len, d_model)`

        # Add dropout
        x = self.dropout(x)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x)

        return x  # Shape `(batch_size, seq_len, d_model)`


class DecoderLayer(tf.keras.layers.Layer):
    def __init__(self,
                 *,
                 d_model,
                 num_heads,
                 dff,
                 dropout_rate=0.1):
        super(DecoderLayer, self).__init__()

        self.causal_self_attention = CausalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)

        self.cross_attention = CrossAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)

        self.ffn = FeedForward(d_model, dff)

    def call(self, x, context):
        x = self.causal_self_attention(x=x)
        x = self.cross_attention(x=x, context=context)

        # Cache the last attention scores for plotting later
        self.last_attn_scores = self.cross_attention.last_attn_scores

        x = self.ffn(x)  # Shape `(batch_size, seq_len, d_model)`
        return x


class Decoder(tf.keras.layers.Layer):
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size,
                 dropout_rate=0.1):
        super(Decoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.pos_embedding = PositionalEmbedding(vocab_size=vocab_size,
                                                 d_model=d_model)
        self.dropout = tf.keras.layers.Dropout(dropout_rate)
        self.dec_layers = [
            DecoderLayer(d_model=d_model, num_heads=num_heads,
                         dff=dff, dropout_rate=dropout_rate)
            for _ in range(num_layers)]

        self.last_attn_scores = None

    def call(self, x, context):
        # `x` is token-IDs shape (batch, target_seq_len)
        x = self.pos_embedding(x)  # (batch_size, target_seq_len, d_model)

        x = self.dropout(x)

        for i in range(self.num_layers):
            x  = self.dec_layers[i](x, context)

        self.last_attn_scores = self.dec_layers[-1].last_attn_scores

        # The shape of x is (batch_size, target_seq_len, d_model)
        return x


class Transformer(tf.keras.Model):
    def __init__(
        self,
        *,
        num_layers: int = 4,
        d_model: int = 128,
        num_heads: int = 8,
        dff: int = 512,
        input_vocab_size: Sequence[int],
        target_vocab_size: Sequence[int],
        dropout_rate: int = 0.1,
    ):
        super().__init__()
        self.encoder = Encoder(num_layers=num_layers, d_model=d_model,
                               num_heads=num_heads, dff=dff,
                               vocab_size=input_vocab_size,
                               dropout_rate=dropout_rate)

        self.decoder = Decoder(num_layers=num_layers, d_model=d_model,
                               num_heads=num_heads, dff=dff,
                               vocab_size=target_vocab_size,
                               dropout_rate=dropout_rate)

        self.final_layer = tf.keras.layers.Dense(target_vocab_size)

    def call(self, inputs):
        # To use a Keras model with `.fit` you must pass all your inputs in the
        # first argument
        context, x  = inputs

        context = self.encoder(context)  # (batch_size, context_len, d_model)

        x = self.decoder(x, context)  # (batch_size, target_len, d_model)

        # Final linear layer output
        logits = self.final_layer(x)  # (batch_size, target_len, target_vocab_size)

        try:
            # Drop the keras mask, so it doesn't scale the losses/metrics
            # b/250038731
            del logits._keras_mask
        except AttributeError:
            pass

        # Return the final output and the attention weights
        return logits

    @classmethod
    def from_pretrained(
        cls,
        *,
        num_layers: int = 4,
        d_model: int = 128,
        num_heads: int = 8,
        dff: int = 512,
        input_vocab_size: Sequence[int],
        target_vocab_size: Sequence[int],
        dropout_rate: int = 0.1,
        hf_hub_repo_id: str = "phunc20/uncased_tf_tuto_transformer",
    ):
        transformer = Transformer(
            num_layers=num_layers,
            d_model=d_model,
            num_heads=num_heads,
            dff=dff,
            input_vocab_size=input_vocab_size,
            target_vocab_size=target_vocab_size,
            dropout_rate=dropout_rate,
        )
        transformer_local_dir = Path(snapshot_download(hf_hub_repo_id))
        basename = hf_hub_repo_id.split("/")[-1]
        transformer.load_weights(transformer_local_dir/basename)

        return transformer


class Translator(tf.Module):
    def __init__(self, *, tokenizers, transformer):
        self.tokenizers = tokenizers
        self.transformer = transformer

    def __call__(self, sentence, max_length=128):
        # Implementation 1: Remainders
        #tokens = self.tokenizers.vi.lookup(output)[0]

        ## `tf.function` prevents us from using the attention_weights that were
        ## calculated on the last iteration of the loop.
        ## So, recalculate them outside the loop.
        #self.transformer([encoder_input, output[:,:-1]], training=False)
        #attention_weights = self.transformer.decoder.last_attn_scores
        #return text, tokens, attention_weights

        # Implementation 2: 128 tokens after 128
        start_end = self.tokenizers.vi.tokenize([''])[0]
        start = start_end[0][tf.newaxis]
        end = start_end[1][tf.newaxis]

        assert isinstance(sentence, tf.Tensor)
        if len(sentence.shape) == 0:
            sentence = sentence[tf.newaxis]
        sentence = self.tokenizers.vina.tokenize(sentence).to_tensor()
        n_tokens = sentence.shape.as_list()[-1]
        pieces = []
        for begin_index in tf.range(0, n_tokens, max_length):
            encoder_input = sentence[:, begin_index:begin_index+max_length]
            output_array = tf.TensorArray(dtype=tf.int64, size=0, dynamic_size=True)
            if begin_index == 0:
                output_array = output_array.write(0, start)
            else:
                output_array = output_array.write(0, sentence[:, begin_index-1])

            for i in tf.range(max_length):
                output = tf.transpose(output_array.stack())
                predictions = self.transformer([encoder_input, output], training=False)
                predictions = predictions[:, -1:, :]  # Shape `(batch_size, 1, vocab_size)`
                predicted_id = tf.argmax(predictions, axis=-1)
                output_array = output_array.write(i+1, predicted_id[0])
                if predicted_id == end:
                    break

            output = tf.transpose(output_array.stack())
            text_tensor = self.tokenizers.vi.detokenize(output)[0]  # Shape: `()`
            pieces.append(text_tensor.numpy().decode("utf-8"))

        text = "".join(pieces)
        return text

    def translate(self, s: str):
        s = uncased_vina_normalizer.normalize_str(s)
        s = tf.constant(s)
        # Implementation 1: Remainder
        #text_tensor, _, _ = self(s)
        text = self(s)
        text = BertDetokenizeHelper.restore_decimal_point(text)
        text = BertDetokenizeHelper.restore_punctuation(text)
        return text

    @classmethod
    def from_pretrained(
        cls,
    ):
        tokenizers = Tokenizers.from_pretrained()
        transformer = Transformer.from_pretrained(
            input_vocab_size=tokenizers.vina.get_vocab_size().numpy().tolist(),
            target_vocab_size=tokenizers.vi.get_vocab_size().numpy().tolist(),
        )
        return cls(tokenizers=tokenizers, transformer=transformer)
